require 'time'
require 'json'
require 'uu/os/lang/json_patch'
require 'uu/os/lang/exception_ext'
require 'uu/os/ipc/error_kind'
require 'uu/os/ipc/error_codes'

module UU
  module OS
    module IPC

      # Ancestor of all errors caused by inter-process communication.
      class IPCError < StandardError

        # Kind of error (see {UU::OS::IPC::ErrorKind} for possible values).
        # @return [String] Kind of error.
        attr_reader :kind

        # Vendor defined error code.
        # @return [String] Error code.
        attr_reader :code

        # Additional error attributes.
        # @return [Hash] Error attributes.
        attr_reader :attributes

        # Flag if exception was raised locally or
        # is result of transparently deserialized
        # server fault.
        # @return [TrueClass, FalseClass]
        # @private
        attr_accessor :remote # :private:

        # TODO APPLICATION package in connection layer, really ? move to application layer
        # @private
        SCHEMA_URI = 'ues:SYSTEM:UU.OS.APPLICATION/ERROR_MESSAGE-SCHEMA-V3'

        # Creates new instance of IPCError. If {UU::OS::IPC::ErrorKind kind} is not given,
        # error is created as {UU::OS::IPC::ErrorKind::SERVER SERVER}.
        # @param [Array<Object>] args Instantiation parameters. Possible values are:
        #
        #   * code [String], message [String]
        #   * code [String], attributes [Hash]
        #   * code [String], cause [Exception]
        #   * kind [String], code [String], message [String]
        #   * kind [String], code [String], attributes [Hash]
        #   * kind [String], code [String], cause [Exception]
        #   * kind [String], code [String], message [String], attributes [Hash]
        #   * kind [String], code [String], message [String], cause [Exception]
        #   * kind [String], code [String], message [String], attributes [Hash], cause [Exception]
        #   * data [String, Hash] (To be used only for instantiation of error from remote response!)
        def initialize(*args)
          @kind = nil
          @code = nil
          @message = nil
          @attributes = nil
          @detail_messages = nil
          @remote = false
          if args.size == 1
            @remote = true
            if args[0].kind_of?(Hash)
              initialize_from_hash(args[0])
            else
              begin
                initialize_from_json(args[0])
              rescue JSON::ParserError
                @message = args[0]
              end
            end
          elsif args.size == 2
            @code = args[0].to_s if !args[0].nil?
            if args[1].kind_of?(Hash)
              @attributes = args[1].dup
            elsif args[1].kind_of?(Exception)
              process_cause(args[1])
            else
              @message = args[1].to_s if !args[1].nil?
            end
          elsif args.size == 3
            # In case of three arguments, we can have either
            # [kind, code, attributes] or [code, message, attributes].
            # The only way to distinguish these options (because type
            # is [String, String, Hash] for both cases) is by checking
            # value of first argument.
            kind_defined = false
            UU::OS::IPC::ErrorKind.constants.each do |kind|
              if UU::OS::IPC::ErrorKind.const_get(kind) == args[0]
                kind_defined = true
                break
              end
            end
            if kind_defined
              @kind = args[0].to_s if !args[0].nil?
              @code = args[1].to_s if !args[1].nil?
            else
              @code = args[0].to_s if !args[0].nil?
              @message = args[1].to_s if !args[1].nil?
            end
            if args[2].kind_of?(Hash)
              @attributes = args[2].dup
            elsif args[2].kind_of?(Exception)
              process_cause(args[2])
            else
              @message = args[2].to_s if !args[2].nil?
            end
          elsif args.size == 4
            @kind = args[0].to_s if !args[0].nil?
            @code = args[1].to_s if !args[1].nil?
            @message = args[2].to_s if !args[2].nil?
            if args[3].kind_of?(Hash)
              @attributes = args[3].dup if !args[3].nil?
            elsif args[3].kind_of?(Exception)
              process_cause(args[3])
            else
              raise ArgumentError, "no implicit conversion of #{args[3].class} to Hash or Exception" if !args[3].nil?
            end
          elsif args.size == 5
            raise ArgumentError, "no implicit conversion of #{args[3].class} to Hash" if !args[3].nil? && !args[3].kind_of?(Hash)
            raise ArgumentError, "no implicit conversion of #{args[4].class} to Exception" if !args[4].nil? && !args[4].kind_of?(Exception)
            @kind = args[0].to_s if !args[0].nil?
            @code = args[1].to_s if !args[1].nil?
            @message = args[2].to_s if !args[2].nil?
            @attributes = args[3].dup if !args[3].nil?
            process_cause(args[4])
          elsif args.size != 0
            raise ArgumentError, "wrong number of arguments (#{args.size} for 0..5)"
          end
          @kind ||= UU::OS::IPC::ErrorKind::SERVER
          @code ||= UU::OS::IPC::ErrorCodes::UNEXPECTED_PROBLEM
          # Call ancestor constructor with default message
          super(format_message(@code, @message, @attributes))
        end

        # Adds additional message to error.
        # @param [String] code Message code.
        # @param [String] message Message text.
        # @param [Hash] attributes Additional message attributes.
        # @return [void]
        def add_detail_message(code, message, attributes = {})
          @detail_messages ||= []
          if code.kind_of?(Hash)
            msg = code
          else
            msg = {code: code, message: message, attributes: attributes.dup}
          end
          msg[:code] ||= UU::OS::IPC::ErrorCodes::UNEXPECTED_PROBLEM
          @detail_messages << msg
        end

        # Returns formatted message (including detail messages).
        # @return [String] Formatted message.
        def message
          tmp = format_message(@code, @message, @attributes)
          if !@detail_messages.nil? && !@detail_messages.empty?
            @detail_messages.each do |msg|
              tmp << "\n" << format_message(msg[:code], msg[:message], msg[:attributes])
            end
          end
          return tmp
        end

        # Returns main exception message in raw (unformatted) form.
        # @return [Hash] Raw message with +:code+, +:message+ and
        #    +:attributes+.
        def main_message
          return {code: @code, message: @message, attributes: @attributes}
        end

        # Returns list of detail messages in raw (unformatted) form.
        # @return [List<Hash>] List of raw messages where each message
        #    has +:code+, +:message+ and +:attributes+.
        def detail_messages
          return @detail_messages
        end

        protected

        # @see Exception#to_hash
        # @private
        def to_hash # :nodoc:
          hash = {}
          hash[:schemaUri] = UU::OS::IPC::IPCError::SCHEMA_URI
          hash.merge! super
          hash[:kind] = @remote ? UU::OS::IPC::ErrorKind::SERVER : @kind
          hash[:kind] ||= UU::OS::IPC::ErrorKind::SERVER
          hash[:code] = @code
          hash[:code] ||= UU::OS::IPC::ErrorCodes::UNEXPECTED_PROBLEM
          hash[:attributes] = @attributes
          hash[:message] = @message
          hash[:detailMessages] = @detail_messages
          # TODO Add "trace" part
          return hash
        end

        private

        # Initializes exception from JSON data.
        # @private
        def initialize_from_json(json) # :nodoc:
          hash = JSON.parse(json, symbolize_names: true)
          initialize_from_hash(hash)
        end

        # Initializes exception from Hash data.
        # @private
        def initialize_from_hash(hash) # :nodoc:
          # Must support new and also "legacy" exception structure.
          @id = hash[:id]
          @timestamp = Time.parse(hash[:timestamp]) if !hash[:timestamp].nil? && !hash[:timestamp].empty?
          @kind = hash[:kind]
          if @kind.nil? && hash.has_key?(:status)
            status = hash[:status].to_i
            if status >= 500
              @kind = UU::OS::IPC::ErrorKind::SERVER
            else
              @kind = UU::OS::IPC::ErrorKind::CLIENT
            end
          end
          @code = hash[:code] || legacy_error_code(hash)
          if hash.has_key?(:errorMessages)
            # Legacy structure of attributes and messages
            @attributes = {}
            if hash.has_key?(:attributes)
              hash[:attributes].each do |attribute|
                # Attribute is composed of name and value attributes
                name = attribute.delete(:name)
                value = attribute.values[0]
                @attributes[name.to_sym] = value
              end
            end
            legacy_messages = hash[:errorMessages]
            default_message = transform_legacy_message(legacy_messages.delete_at(0))
            @message = default_message[:message]
            @attributes.merge!(default_message[:attributes])
            @detail_messages = [] if legacy_messages.size > 0
            legacy_messages.each do |legacy_message|
              @detail_messages << transform_legacy_message(legacy_message)
            end
          else
            @attributes = hash[:attributes]
            @message = hash[:message]
            @detail_messages = hash[:detailMessages]
          end
          # TODO Process "trace" part
        end

        # Formats message.
        # @param [String] code Message code.
        # @param [String] message Message text.
        # @param [Hash] attributes Message attributes.
        # @return [String] Formatted message.
        # @private
        def format_message(code, message, attributes = nil) # :nodoc:
          result =  "#{code}: #{message}"
          return result if attributes.nil? || attributes.empty?
          result.dup.scan(/#\{([a-zA-Z0-9_\-]+)\}/) do |match|
            key = match[0].to_sym
            val = attributes[key]
            result.gsub!("\#{#{key}}", val.to_s) if !val.nil?
          end
          return result
        end

        # Transforms legacy exception message data format to current
        # message data format.
        # @private
        def transform_legacy_message(message) # :nodoc:
          result = {}
          result[:code] = message[:code] || legacy_error_code(message) || 'NA'
          text = (message[:message] || message[:localizedMessage]).to_s
          text.dup.scan(/\{([0-9]+)\}/) do |match|
            text.gsub!("{#{match[0]}}", "\#{#{match[0]}}")
          end
          result[:message] = text
          attributes = {}
          if message.has_key?(:parameters)
            message[:parameters].each_with_index do |parameter, idx|
              attributes[idx.to_s.to_sym] = parameter
            end
          end
          result[:attributes] = attributes
          return result
        end

        # Transforms legacy error code data format to current
        # error code data format.
        # @private
        def legacy_error_code(message) # :nodoc:
          return nil if !message.has_key?(:errorCodeClass) || !message.has_key?(:errorCode)
          prefix = 'UU.OS'
          enum = message[:errorCodeClass].split(/[$.]/)[-1]
          code = message[:errorCode]
          return "#{prefix}/#{enum}.#{code}"
        end

        # Processes cause by adding its messages as detail messages.
        # @private
        def process_cause(cause) # :nodoc:
          return if cause.nil?
          if cause.kind_of?(IPCError)
            add_detail_message(cause.main_message, nil)
            if !cause.detail_messages.nil?
              cause.detail_messages.each do |message|
                add_detail_message(message, nil)
              end
            end
          else
            add_detail_message(nil, cause.message)
          end
        end

      end

    end
  end
end
